#!/usr/bin/python
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2022  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##


import sys
import os
import sqlite3
import datetime

try:
    wapt_root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'

from waptserver.config import DEFAULT_CONFIG_FILE, load_config
#from waptserver.utils import *
from waptserver.utils import logger
#from waptserver.model import *
from waptserver.model import load_db_config
import waptserver.globalsupdate as gl
from waptcrypto import SSLPrivateKey, SSLCRL, SSLCertificate

from huey.api import SqliteHuey
from huey import crontab

from waptserver.common import licences_list


# hackish way to load specific server config inifile from command line when this module is imported by huey consumer.
# the custom config waptserver.ini filename is supposed to be the last argument of the command line, without switch.
if hasattr(gl,'conf') and hasattr(gl,'tasks_config_file'):
    conf = gl.conf
    tasks_config_file = gl.tasks_config_file
elif not 'conf' in globals() or not 'tasks_config_file' in globals():
    # hack to get a config filename from end of command line arguments without rewriting the huey consumer.
    if len(sys.argv) > 2 and os.path.isfile(sys.argv[-1]) and os.path.splitext(sys.argv[-1])[1] == '.ini':
        conf = load_config(sys.argv[-1])
        tasks_config_file = sys.argv[-1]
    elif 'tasks_config_file' in os.environ:
        tasks_config_file = os.environ['tasks_config_file']
        conf = load_config(tasks_config_file)
    else:
        conf = load_config()
        tasks_config_file = DEFAULT_CONFIG_FILE

licences_list.conf = conf

# location of tasks queue shared storage db
#  waptserver enqueue task in these db, and huey consumer dequeues tasks from here
tasks_db_dir = os.path.dirname(conf.get('wapt_huey_db'))
if not os.path.isdir(tasks_db_dir):
    os.makedirs(tasks_db_dir)

if 'huey' not in globals():
    # check here if we should upgrade structure.
    try:
        huey = SqliteHuey('wapt', filename=conf.get('wapt_huey_db'), utc=False)
    except sqlite3.OperationalError:
        # recreate if schema is broken
        # anyway, these are temporary tasks.
        if os.path.isfile(conf.get('wapt_huey_db')):
            os.unlink(conf.get('wapt_huey_db'))
        huey = SqliteHuey('wapt', filename=conf.get('wapt_huey_db'))
    huey.flush_locks()
    load_db_config(conf)

@huey.periodic_task(crontab(minute='0', hour='*/1'), context=True, name='resign_crl')
def resign_crl(task):
    """Check validity of CRL and resign it if it will expire before next round.
    """
    if (conf['clients_signing_key'] and conf['clients_signing_certificate'] and conf['clients_signing_crl'] and
            os.path.isfile(conf['clients_signing_key']) and os.path.isfile(conf['clients_signing_certificate'])):
        signing_key = SSLPrivateKey(conf['clients_signing_key'])
        signing_cert = SSLCertificate(conf['clients_signing_certificate'])
        logger.info('Resigning CRL %s' % conf['clients_signing_crl'])
        try:
            crl = SSLCRL(conf['clients_signing_crl'], cakey=signing_key, cacert=signing_cert)
            if not crl.crl or crl.next_update <= datetime.datetime.utcnow() + datetime.timedelta(hours=1):
                crl.revoke_cert(crl_ttl_days=conf['clients_signing_crl_days'])
            crl.save_as_pem()
        except Exception as e:
            logger.warning('Unable to sign CRL %s: %s' % (conf['clients_signing_crl'], repr(e)))
